///////////////////////////////////////////////////////////////////////////////
//  Modified by Li Teng
//  Date   : 2012-Jan
//  Description : bug fixed for the first step: finding maximal sample set for each genomic loci
//                read prior knowledge to semi-supervise CoSBI
//                merge implemented
//
//  Written by  : Duygu Ucar
//  Date    : Oct - Nov 2009 
//  
//  Description : Given a time series microarray data (M X N X T) - M genes, N conditions, and T time samples,   
//  the algorithm idenfities coherent gene groups over sample sets
//
//  1) Pre-processing: 
//  1-a) The program first calculates a M by M binary matrix for each gene 
//  which indicates if the gene is coherent for the corresponding sample pair, matrix C.
//  1-b) Next, cliques within each of these C matrics are identified. 
//
//  2) output biclusters satisfying giving parameters by tree enumeration
//
///////////////////////////////////////////////////////////////////////////////

// My header files
#include "SS_CoSBI_header.h"

using namespace std;

Array3D	array3d; // holds the 3d data n a single matrix of this order (G,S,T)
int G, S, AT;
float T, C;

void print_C_array(int C[], int size){
  for(int i = 0; i< size ; i++){
    for(int j = 0; j< size ; j++){
      cout<<C[i*size+j]<<",";
    }
    cout<<endl;
  }
}

/*------------------------ Li_S------------------------------*/
void ReadConstraints(float *M, char * filename) {
   int i =0;
   FILE *pFile;
   pFile = fopen(filename, "r");
   if(pFile == NULL) {perror(filename); exit(-1);}
   while(!feof(pFile)){
      fscanf(pFile, "%f", &M[i]);
      i++;
   }
   fclose(pFile);
   //printf(" M[0] = %.2f\t, M[%d] = %.2f\n", M[0], i-1, M[i-1]);
}

/*
    floyd_warshall(), this function updated the distance matrix D
*/
void floyd_warshall( float *D, int n ) {
    int i, j, k;
    float dist[100][100];

    for(i=0;i<n;i++)
       for(j=0;j<n;j++)
          dist[i][j] = D[ i * n + j ];

    for (k = 0; k < n; ++k) {
        for (i = 0; i < n; ++i)
            for (j = 0; j < n; ++j)
                /* If i and j are different nodes and if
                    the paths between i and k and between
                    k and j exist, do */
                if (i != j)
                    /* See if you can't get a shorter path
                        between i and j by interspacing
                        k somewhere along the current
                        path */
                    if (dist[i][k] + dist[k][j] < dist[i][j])
                        dist[i][j] = dist[i][k] + dist[k][j];
    }
    for(i=0;i<n;i++)
       for(j=0;j<n;j++)
          D[ i * n + j ] = dist[i][j];  
}
/*.........................Li_E--------------------------------*/
//
void searchClique(int* cMatrix, vector<int> head, vector<int> tail, vector< vector<int> >* maxCliq, int size, int sizeThr);
void recurSearch(vector<int> head, vector<int> tail, vector< pair< vector<int>, vector<int> > >* gsSet, int minS, int minG, vector< vector<pair<int,int> > >* invList, vector<pair<int,int> > ILintsec);

vector< vector<int> > find_cliques (int* cMatrix, int size, int sizeThr)
{
    vector< vector<int> > maxCliq;
    for(int n=0; n<=size-sizeThr; n++)
    {
        vector<int> head (1, n);
        //get tail
        vector<int> tail;
        for(int m = n+1; m < size; m++)
        {
            tail.push_back(m);
        }
        
        searchClique(cMatrix, head, tail, &maxCliq, size, sizeThr);
    }
    return maxCliq;
}

void searchClique(int* cMatrix, vector<int> head, vector<int> tail, vector< vector<int> >* maxCliq, int size, int sizeThr)
{
            
    //remove elements from tail according to the c matrix
    int i = head[head.size()-1];
    for(int n = 0; n < tail.size(); n++)
    {
        int j = tail[n];
        if (cMatrix[i*size+j] == 0)
        {
            tail.erase(tail.begin()+n);
            n--;    //the index should move one step back after an element is deleted
        }
    }
         
    //check the total length
    if(head.size()+tail.size() < sizeThr)
        return;
   
    // check if the union has already been discovered
    vector<int> continent;
    for(int n = 0; n < head.size(); n++)
    {
        continent.push_back(head[n]);
    }
    for(int n = 0; n < tail.size(); n++)
    {
        continent.push_back(tail[n]);
    }
    for(int n=0; n<(*maxCliq).size(); n++){
       if(includes ( ((*maxCliq)[n]).begin(), ((*maxCliq)[n]).end(), continent.begin(), continent.end() ) ){
//           cout<<"Prune due to redundance..."<<endl;
          return;
       }
    }

    //the situation of |tail|=0
    if(tail.size() == 0)
    {
        (*maxCliq).push_back(head);
    }
    else
    {
        while(tail.size() != 0)
        {
            vector<int> headTemp = head;
            headTemp.push_back(tail[0]);
            tail.erase(tail.begin());
            searchClique(cMatrix, headTemp, tail, maxCliq, size, sizeThr);
        }
    }
    return;
}

vector< pair< vector<int>, vector<int> > > SampleGeneSearch(int size, int minS, int minG,vector< vector<pair<int,int> > >* invList)
{
    vector< pair< vector<int>, vector<int> > > gsSet;
    for (int n=0; n<=size-minS; n++)
    {
        vector<int> head (1, n);

        //get tail
        vector<int> tail;
        for (int m = n+1; m < size; m++)
        {
            if ((*invList)[m].size() >= minG)

                tail.push_back(m);
        }
        recurSearch(head, tail, &gsSet, minS, minG, invList, (*invList)[n]);
    }
    return gsSet;
}
    
void recurSearch(vector<int> head, vector<int> tail, vector< pair< vector<int>, vector<int> > >* gsSet, int minS, int minG, vector< vector<pair<int,int> > >* invList, vector<pair<int,int> > ILintsec)
{
    vector<pair<int,int> > g ((*invList)[head[0]]);
    
    //get the intersection
    int n = head.size()-1;
    //vector<pair<int,int> > tempG (g);
    //g.resize(min(tempG.size(), (*invList)[head[n]].size()));
    g.resize(set_intersection(ILintsec.begin(), ILintsec.end(), (*invList)[head[n]].begin(), (*invList)[head[n]].end(), g.begin())-g.begin());
    
    vector<int> gSet;
    for(int n=0; n<g.size(); n++)
    {
        gSet.push_back(g[n].first);
    }
    gSet.resize(unique(gSet.begin(), gSet.end())-gSet.begin());

    if(gSet.size()<minG)
        return;

    //check the total length
    if(head.size()+tail.size() < minS)
        return;
        
    while(tail.size() != 0)
    {
        vector<int> headTemp = head;
        headTemp.push_back(tail[0]);
        tail.erase(tail.begin());
        recurSearch(headTemp, tail, gsSet, minS, minG, invList, g);
    }
    
    if(head.size()>=minS)
    {
        for(int n=0; n<(*gsSet).size(); n++)
        {
            if((*gsSet)[n].first.size()==gSet.size() && (*gsSet)[n].second.size()>=head.size())
            {
                //if(includes((*gsSet)[n].first.begin(), (*gsSet)[n].first.end(), gSet.begin(), gSet.end()))
                {
                    if(includes((*gsSet)[n].second.begin(), (*gsSet)[n].second.end(), head.begin(), head.end()))
                    {
                        return;
                    }
                }
            }
            else if((*gsSet)[n].first.size()==gSet.size() && (*gsSet)[n].second.size()<head.size())
            {
                //if(includes(gSet.begin(), gSet.end(), (*gsSet)[n].first.begin(), (*gsSet)[n].first.end()))
                {
                    if(includes(head.begin(), head.end(), (*gsSet)[n].second.begin(), (*gsSet)[n].second.end()))
                    {
                        (*gsSet).erase((*gsSet).begin()+n);
                        n--;
                    }
                }
            }
        }
        
	pair< vector<int>, vector<int> > gs (gSet, head);
	// if for any two genes there is a sample that the correlation between these two genes on that sample is larger than C then OK, otherwise prune

	for(int i = 1; i < gSet.size(); i++)
        {
            for(int k = 0; k < i; k++)            
            {
                for(int j = 0; j < head.size(); j++)
		{
                    if(cross_cor(array3d.get3d(gSet[i],head[j]), array3d.get3d(gSet[k],head[j]), array3d.T, array3d.T) >= C)
		    //if(pearsoncorrelation(array3d.get3d(gSet[i],head[j]), array3d.get3d(gSet[k],head[j]), array3d.T) >= C)
                    goto GoOn;
		 }
		 return;
                 GoOn:
		  ;
	    }
	 }
	 //  pair< vector<int>, vector<int> > gs (gSet, head);
	 (*gsSet).push_back(gs);
    }
}

/*----------------------- Merge highly overlapped biclusters ------------------------------------*/
vector< pair< vector<int>, vector<int> > >  MergeClusters( vector< pair< vector<int>, vector<int> > > gsSet, float Threshold)
{
   int n; //number of original biclusters

   n = gsSet.size(); // number of current Nodes

   vector<int> flag, CurrentNode, Node1, Node2;
   vector< vector<float> > OverlapMatrix;
   vector<float> NodeOverlap;
   vector< vector<int> > NodeIndex;
      
   for(int i=0; i< n ; i++){
       flag.push_back(1);
       CurrentNode.push_back(i);
       NodeIndex.push_back(CurrentNode);

       CurrentNode.pop_back();
    }    
   
     vector <int> gset1, gset2, hset1, hset2;
     int ng1, ng2, nh1,nh2, gunique, hunique; 
     int clusterid1, clusterid2;
     float overlap;
 
    // construct the initial overlapping matrix
    for(int i =0; i< n;i++){
       NodeOverlap.resize(0);
       for(int j = 0; j<n;j++){
             if(i==j){
                NodeOverlap.push_back(-1);
             }
             else{
        // calculate overlapping between two biclusters
             Node1 = NodeIndex.at(i);
             Node2 = NodeIndex.at(j);
        //cout <<"size of cluster1 = "<< cluster1.size()<<", size of cluster2 = "<<cluster2.size()<<endl;
             clusterid1 = Node1.at(0);
             clusterid2 = Node2.at(0);
             gset1 = gsSet.at(clusterid1).first;
             hset1 = gsSet.at(clusterid1).second;

             gset2 = gsSet.at(clusterid2).first;
             hset2 = gsSet.at(clusterid2).second;

             ng1 = gset1.size(); ng2 = gset2.size();
             nh1 = hset1.size(); nh2 = hset2.size();

             for(int q = 0; q<gset2.size(); q++){
                gset1.push_back(gset2[q]);
             }
             for(int q = 0; q<hset2.size(); q++){
                hset1.push_back(hset2[q]);
             }
 
             sort(gset1.begin(),gset1.end());
	     sort(hset1.begin(),hset1.end());
             gunique = unique(gset1.begin(), gset1.end()) - gset1.begin();
             hunique = unique(hset1.begin(), hset1.end()) - hset1.begin();
                          
             overlap =   ( float ( ng1 + ng2 - gunique) * float  (nh1 + nh2 - hunique) ) / float ( gunique * hunique)  ;
             //cout<<"bicluster "<<i<<" and "<< j << "overlap "<<overlap<<endl;
             NodeOverlap.push_back(overlap);
             }
          }
      OverlapMatrix.push_back(NodeOverlap);
     }

     int mergeindex1, mergeindex2;
     float bigoverlap, withinoverlap;
     int flagMerge = 1;
     while(flagMerge){
        bigoverlap = 0;

        for(int i =0; i< n-1;i++){ //find two nodes which have the highest overlap
           if( flag.at(i) == 1) {
 
              for(int j = i+1; j<n;j++){
                 if( flag.at(j) == 1) {
                    overlap = OverlapMatrix.at(i).at(j);
                    if (overlap >bigoverlap){
                       bigoverlap = overlap;
                       mergeindex1 = i;
                       mergeindex2 = j;
                    }
                 }
              }
           }
        }

//        cout<< " bigoverlap = "<< bigoverlap<<endl;
        if(bigoverlap >= Threshold){  //merge two biclusters
//            cout <<"Merge Nodes "<< mergeindex1 <<" and "<< mergeindex2 <<endl;
           flag.at(mergeindex1) = 0;
           flag.at(mergeindex2) = 0;
        
           Node1 = NodeIndex.at(mergeindex1);
           Node2 = NodeIndex.at(mergeindex2);
            
           for(int i = 0; i<Node2.size();i++){
              Node1.push_back( Node2.at(i) );
           }
           NodeIndex.push_back ( Node1 );
           //calculate the overlapping ration of all the node to the new node
           NodeOverlap.resize(0);
           for(int i= 0; i<n ;i++){
              if(flag.at(i) == 1){
                 Node2 = NodeIndex.at(i); //Node2 is an old node, Node1 is the new node
                 withinoverlap = 1;
                 for(int j=0;j<Node2.size();j++)
                    for(int k=0; k<Node1.size();k++){
                       clusterid1 = Node2.at(j);
                       clusterid2 = Node1.at(k);
//                        cout <<clusterid1 <<"\t"<<clusterid2<<endl;
                       if(OverlapMatrix.at( clusterid1 ).at( clusterid2  ) < withinoverlap)
                          withinoverlap = OverlapMatrix.at(clusterid1).at(clusterid2);
                    }
                 OverlapMatrix.at(i).push_back(withinoverlap);
//                  cout<<"Node "<<i<<" and the new node overlap "<<withinoverlap<<endl;
                 NodeOverlap.push_back(withinoverlap);
             }
             else{
                NodeOverlap.push_back(-1);
             }   
           } 
           NodeOverlap.push_back(-1);
           OverlapMatrix.push_back(NodeOverlap);
           n = n + 1;
           flag.push_back(1);

        }
        else{
           flagMerge = 0; // Merge finish
        }
     }

     // construct the genelist and histone list for each merged bicluster
    vector <int> gset, hset;
    vector< pair< vector<int>, vector<int> > > gsSetNew;
    for(int i = 0; i<n;i++){
       if(flag.at(i) == 1){
          CurrentNode = NodeIndex.at(i);
          gset.resize(0);
          hset.resize(0);
          for(int j = 0;j<CurrentNode.size(); j++){
             clusterid1 = CurrentNode.at(j);
             gset.insert( gset.end(), gsSet.at(clusterid1).first.begin(),gsSet.at(clusterid1).first.end());
             hset.insert( hset.end(), gsSet.at(clusterid1).second.begin(),gsSet.at(clusterid1).second.end());
//              cout <<" add " <<gsSet.at(clusterid1).first.size() <<" made to "<<gset.size()<<endl;
          }
          sort(gset.begin(), gset.end());
          sort(hset.begin(), hset.end());
          gset.resize(unique(gset.begin(), gset.end()) - gset.begin());
          hset.resize(unique(hset.begin(), hset.end()) - hset.begin());
          gsSetNew.push_back(make_pair( gset, hset));
       }
    }
   return(gsSetNew);
}

int main(int argc, char *argv[])
{
        char	filename[100], Mfilename[100];
        Mfilename[0] = '\0';
	ofstream out;

        if(argc < 5) {
	  cout << endl << "Usage: " << argv[0] << " -fFile [other options]" << endl;
	  cout << "==================== OPTIONS =====================" << endl;
	  cout << "-f     -ffilename  File for input data	   "   << endl;
	  cout << "-g     -g100     Minimal Gene Size              "   << endl;
	  cout << "-s     -s3       Minimal Sample Size            "         << endl;
	  cout << "-t     -t0.75    Minimal Within Gene Correlation "         << endl;
	  cout << "-c     -c0.65    Minimal Within Sample Correlation"         << endl<< endl;
	  cout << "-m               (Optional) Run original CoSBI without -m" <<endl;
          cout << "       -mfilename File for Must-link constraints"   << endl<<endl;
	  cout << "==================== EXAMPLES =====================" << endl;
	  cout << "Eg, ./SS_CoSBI -finput.tricluster -g1000 -s3 -t0.7 -c0.5 -mM39.txt" << endl;
          cout << "Or, ./SS_CoSBI -finput.tricluster -g1000 -s3 -t0.7 -c0.5" << endl;
	  cout << "==================================================" << endl << endl;
	  exit(1);
        }
	
        for(int i=1; i<argc; i++){
	  if( argv[i][0] == '-' )
	    switch( argv[i][1] ){
	    case 'f':  	
              strcpy( filename, argv[i]+2 );        break;
	    case 'g':
	      G = (int) atoi(argv[i]+2); break;
	    case 's':
	      S = (int) atoi(argv[i]+2); break;
	    case 't':
	      T = (float) atof(argv[i]+2); break;
	    case 'c':
	      C = (float) atof(argv[i]+2); break;
	    case 'm':
	      strcpy(Mfilename, argv[i]+2); break;
	    default: cout << "Error Input Options: " << argv[i] << endl; exit(0); break;
	    }
	  else{	cout << "Error Input Options: " << argv[i] << endl;  exit(0); }
        }

        // start clock
	time_t start,end;
	double dif;

	time (&start);
	// Define variables
	//Array3D	array3d; // holds the 3d data n a single matrix of this order (G,S,T)
	Input	in(filename, array3d); 
	in.readTabFile(); // my data is loaded into the array3d matrix

        float threshold = T; // to convert the correlation scores into a binary value

	int** matrix; // This will hold the Cross Correlation matrices for each gene
	int numCols = array3d.G;
	int numRows = array3d.S*array3d.S;
   
                
        //Li-S
        float *MM, *CC, *DD; // This will hold the original&updated Cross Correlation Matrix and Distance Matrix //Li

	/* Initialize  2d array */
	/* Allocate pointer memory for the first dimension of the matrix[][]; */
	matrix = (int **) malloc(numCols * sizeof(int *));
	if(NULL == matrix){free(matrix); printf("Memory allocation failed while allocating for matrix[].\n"); exit(-1);}
	
	/* Allocate integer memory for the second dimension of the matrix[][]; */
	for(int x = 0; x < numCols; x++){
	  matrix[x] = (int *) malloc(numRows * sizeof(int));
	  if(NULL == matrix[x]){free(matrix[x]); printf("Memory allocation failed while allocating for matrix[x][].\n"); exit(-1);}
	}
	
	for(int y = 0; y < numRows; y++)
	  for(int x = 0; x < numCols; x++)
	    matrix[x][y] = 0;
	
        //** 1. Read or Generate the Must-Link Constraint Matrix M
        MM = (float *) malloc (array3d.S * array3d.S * sizeof(float));
        if(MM == NULL) {free(MM); printf("Memory allocation failed while allocating for MM.\n"); exit(-1);}        

        if(Mfilename[0] == '\0'){
            cout<<"Running original CoSBI..."<<endl;
            for(int x = 0; x < array3d.S; x++)
                for(int y = 0; y < array3d.S; y++){
                    MM[ x * array3d.S + y] = 0;     // test with MMij = 0;
                }
        }
        else {
           cout <<"Running SS_CoSBI..."<<endl;
           ReadConstraints(MM, Mfilename);
        }
        
        //** initalize the cross correlation matrix CC and Distance matrix DD
        CC = (float *) malloc (array3d.S * array3d.S * sizeof(float));
        if(CC == NULL) {free(CC); printf("Memory allocation failed while allocating for CC.\n"); exit(-1);}

        DD = (float *) malloc (array3d.S * array3d.S * sizeof(float));
        if(DD == NULL) {free(DD); printf("Memory allocation failed while allocating for DD.\n"); exit(-1);}

        for(int x =0;x < array3d.S;x++)
           for(int y=0;y<array3d.S;y++){
              CC[x * array3d.S + y] = 0;
              DD[x * array3d.S + y] = 0;
           }
         //Li-E

	// For each gene construct a S x S matrix
	for(int i = 0; i < (array3d.G); i++){
	  for(int j = 0; j <array3d.S ; j++){
	    for(int k = 0 ; k <array3d.S; k++){
//               cout<<i<<" "<<j<<" " <<k<<endl;
//               getchar();
	      float temp = cross_cor( array3d.get3d(i,j), array3d.get3d(i,k), array3d.T, array3d.T);
	      //float temp = pearsoncorrelation(array3d.get3d(i,j), array3d.get3d(i,k), array3d.T);
              CC[j*array3d.S + k] = temp;              
	    }
	  }

          //** DD = 1 - CC;   //transform Cross Correlation Matrix to Distance Matrix D
          for(int x =0;x < array3d.S;x++)
           for(int y=0;y<array3d.S;y++){
              DD[x * array3d.S + y] = 1 - CC[x * array3d.S + y];
           }
// 
          //** DD = adjust(DD, M); // adjust Distance Matrix D according to the must-link constraint matrix M. 1, Dij = Dij *(1 - Mij), 2, Dij = Dij - Mij
          for(int x =0;x < array3d.S;x++)
           for(int y=0;y<array3d.S;y++){
              DD[x * array3d.S + y] = DD[x * array3d.S + y] * (1 - MM[ x * array3d.S + y ]);  
              //DD[x * array3d.S + y] = DD[x * array3d.S + y] - MM[ x * array3d.S + y] ;
              //if (DD[x * array3d.S + y]<0) DD[x * array3d.S + y] = 0;
           }  

          if(Mfilename[0] == '\0'){    }
          else{
              // update the Distance Matrix with all-pair-shortest-path algorithm, decrease the distance betweeen nodes involved shortest paths with update
              floyd_warshall(DD, array3d.S);  //all-pair-shortest-path algorithm
          }

          // CC = 1 - DD; transform Distance Matrix back to Cross Correlation Matrix. Some of the items are decreased
          for(int x =0;x < array3d.S;x++)
           for(int y=0;y<array3d.S;y++){
              CC[x * array3d.S + y] = 1 - DD[x * array3d.S + y];
           }
//
	  for(int j = 0; j <array3d.S ; j++){
	    for(int k = 0 ; k <array3d.S; k++){
               if ( CC[j * array3d.S + k ] >= threshold){
	          matrix[i][(j*(array3d.S)+k)] = 1;
	       }
            }
          }
  	}

    vector< vector<int> > maxCliq[(const int)array3d.G];
    for(int i = 0; i < (array3d.G); i++)
    {
        //--find_cliques(C);
        maxCliq[i] = find_cliques(matrix[i], array3d.S, S);
        //Display each set in a line
        //cout<<"g"<<i<<": ";
        for(int n = 0; n < maxCliq[i].size(); n++)
        {
            for(int m = 0; m < maxCliq[i].at(n).size(); m++)
            {
                int s = (maxCliq[i]).at(n).at(m);
            }
        }
    }
       
    vector< vector<pair<int,int> > > invList;
    for(int eachS = 0; eachS < array3d.S; eachS++)
    {
        vector<pair<int,int> > temp;
        for(int i = 0; i < (array3d.G); i++)
        {
            for(int n = 0; n < maxCliq[i].size(); n++)
            {
                if (binary_search ((maxCliq[i]).at(n).begin(), (maxCliq[i]).at(n).end(), eachS))
                {
                    temp.push_back(make_pair (i,n));
                }
            }
        }
        invList.push_back(temp);
    }    

    //SampleGeneSearch(39,3,1000,&invList);
    vector< pair< vector<int>, vector<int> > > gsSet = SampleGeneSearch(array3d.S,S,G,&invList);

//========== Check for overlapping and merge the largely overlapped biclusters==============//
    gsSet = MergeClusters(gsSet, 0.75);  // The merging threshold is hard-coded as 0.75

    //Print results to a file
  
     FILE *bS;
     bS = fopen("biclusters.txt","w");
     if(bS == NULL) {perror("biclusters.txt");exit(-1);}
     for (int n=0; n<gsSet.size(); n++)
     {
         for(int i=0; i<gsSet[n].first.size(); i++)
         {
           if(i == gsSet[n].first.size() -1 ){
              fprintf(bS,"%d\n",gsSet[n].first.at(i));
           }
           else{
              fprintf(bS,"%d ",gsSet[n].first.at(i));
           }
         }
         for(int i=0; i<gsSet[n].second.size(); i++)
         {
             if( i == gsSet[n].second.size() - 1){
                fprintf(bS,"%d\n",gsSet[n].second.at(i));
             }
             else{
                fprintf(bS,"%d ",gsSet[n].second.at(i));
             }
         }
     }
     fclose(bS);
     cout<<"Please check file biclusters.txt for result!"<<endl;

/*    time (&end);
    dif = (double)(difftime (end,start)/ (double) 60);
    printf ("Elapsed time for generating GS sets: %.2lf minutes.\n", dif );*/

    return 0;
}


